"""
@brief: agent for RL
@author: Zhengyang Lyu
@date: 2022.9.1
"""

from model import PTM, Encoder, Decoder, D
import torch
import torch.nn as nn
from torch.distributions.categorical import Categorical
from env.env import RSMTEnv

import pdb

# TODO: finish the mask
# leave the mask provided by the environment
# TODO: make the vectorized version
class Actor(nn.Module):
  def __init__(self, dim_e=D, dim_q=360) -> None:
    super(Actor, self).__init__()
    self.encoder = Encoder(dim=dim_e, N=3)
    self.decoder = Decoder(dim_e, dim_q)
    self.dim_e = dim_e
    self.dim_q = dim_q

    self._node_e = None # node embeddings

  def get_u0_probs(self, nodes):
    """
    @brief get a probability distribution of u0
    @param nodes: [#batch_size, #num_nodes, 2]
    @return probs [#batch_size, #num_nodes]
    """
    e = self.encoder(nodes)
    self._node_e = e.clone()
    u0_probs = self.decoder._get_u0_probs(e)
    return u0_probs
  
  def get_u_probs(self, nodes, mask_visited=None, mask_unvisited=None):
    e = self.encoder(nodes)
    u_probs = self.decoder.get_u_probs(e, mask_visited, mask_unvisited)
    return u_probs

  def get_ws_probs(self, nodes, mask_visited=None, mask_unvisited=None):
    e = self.encoder(nodes)
    ws_probs = self.decoder.get_ws_probs(e, self.decoder._last_Eu, 
                                          mask_visited, mask_unvisited)
    return ws_probs

class Critic(nn.Module):
  def __init__(self, dim_e=D, dim_c=256) -> None:
    super(Critic, self).__init__()
    self.dim_e = dim_e
    self.dim_c = dim_c
    self.encoder = Encoder(dim=dim_e, N=3)
    self.g = nn.Linear(dim_e, 1, bias=False)
    self.final = nn.Sequential(
      nn.Linear(dim_e, dim_c, bias=True),
      nn.ReLU(),
      nn.Linear(dim_c, 1, bias=True)
    )

  def forward(self, nodes):
    """
    @param nodes: [#batch, #num_nodes, 2] in dtype float
    @return Expection for each batch, [#num_batch]
    """
    e = self.encoder(nodes) # [#batch_size, #num_nodes, D]
    glimpse_w = self.g(torch.tanh(e)).squeeze() # [#batch_size, #num_nodes]
    glimpse_w = torch.softmax(glimpse_w, dim=1).unsqueeze(1) # [#batch_size, 1, #num_nodes]
    glimpse = torch.bmm(glimpse_w, e).squeeze() # [#batch_size, D]
    b = self.final(glimpse).squeeze() # [#batch_size]
    return b

"""
the action_space is of gym.spaces.Discrete
the agent outputs probabilities, use torch.distributions.categorical.Categorical()
"""
class Policy(nn.Module):
  def __init__(self, 
               actor,
               critic,
               obs_space=None, 
               action_space=None,
               actor_optimizer=None,
               critic_optimizer=None,
               lr_actor=2.5e-4,
               lr_critic=2.5e-4,
               weight_decay_actor=5e-4,
               weight_decay_critic=5e-4,
               loss_fn_critic=torch.nn.MSELoss()
               ) -> None:
    super(Policy, self).__init__()
    self.actor = actor
    self.critic = critic
    self.obs_space = obs_space
    self.action_space = action_space

    if actor_optimizer == None:
      self.actor_optimizer = torch.optim.Adam(self.actor.parameters(), lr=lr_actor, weight_decay=weight_decay_actor) 
    if critic_optimizer == None:
      self.critic_optimizer = torch.optim.Adam(self.critic.parameters(), lr=lr_critic, weight_decay=weight_decay_critic) 
    self.loss_fn_critic = loss_fn_critic

    self.log_probs_buf = {
      "u0": None,
      "u": [],
      "ws": []
    }

  def forward(self, x):
    raise NotImplementedError

  def first_act(self, nodes):
    """
    @brief perform action for t == 0, with the query = 0, get u0
    @param nodes: [#num_batch, #num_nodes, 2]
    @return points: [#num_batch], and the corresponding log_prob [#num_batch]
    """
    probs = self.actor.get_u0_probs(nodes)  # [#num_batch, #num_nodes]
    dist = Categorical(probs)
    u0 = dist.sample()
    batch_size = u0.shape[0]
    _last_Eu = []
    for i in range(batch_size):
      _last_Eu.append(self.actor._node_e[i, u0[i]])
    self.actor.decoder._last_Eu = torch.stack(_last_Eu)
    self.actor.decoder._last_Ev = self.actor.decoder._last_Eu.clone()
    self.log_probs_buf['u0'] = dist.log_prob(u0)
    return u0

  def act(self, nodes, mask_visited=None, mask_unvisited=None):
    """
    @brief perform action for t >= 1
    @param nodes: [#num_batch, #num_nodes, 2]
    @param mask_visited/mask_unvisited: [#num_batch, #num_nodes]
    TODO: gather the input into one obs: contains a batch of obs
    @return u/w/s: [#num_batch], and their corresponding log_prob [#num_batch]
    """
    if mask_visited == None and not mask_unvisited == None:
      mask_visited = ~mask_unvisited
    if mask_unvisited == None and not mask_visited == None:
      mask_unvisited = ~mask_visited
    
    num_nodes = nodes.shape[1]
    u_probs = self.actor.get_u_probs(nodes, 
          mask_visited=mask_visited, mask_unvisited=mask_unvisited)
    u_dist = Categorical(u_probs)
    u = u_dist.sample()
    batch_size = u.shape[0]

    _last_Eu = []
    for i in range(batch_size):
      _last_Eu.append(self.actor._node_e[i, u[i]])
    self.actor.decoder._last_Eu = torch.stack(_last_Eu)

    ws_probs = self.actor.get_ws_probs(nodes,
          mask_visited=mask_visited, mask_unvisited=mask_unvisited)
    ws_dist = Categorical(ws_probs)
    _w = ws_dist.sample()

    self.log_probs_buf['u'].append(u_dist.log_prob(u))
    self.log_probs_buf['ws'].append(ws_dist.log_prob(_w))

    s = torch.where(_w < num_nodes, int(0), int(1)).to(_w.device)
    w = torch.where(_w < num_nodes, _w, _w - num_nodes)

    _last_Ew = []
    _last_Ev = []
    _last_Eh = []
    v = []
    h = []
    for i in range(batch_size):
      _last_Ew.append(self.actor._node_e[i, w[i]])
      if s[i] == 0:
        _last_Ev.append(self.actor._node_e[i, u[i]])
        _last_Eh.append(self.actor._node_e[i, w[i]])
        v.append(u[i])
        h.append(w[i])
      else:
        _last_Ev.append(self.actor._node_e[i, w[i]])
        _last_Eh.append(self.actor._node_e[i, u[i]])
        v.append(w[i])
        h.append(u[i])
    self.actor.decoder._last_Ew = torch.stack(_last_Ew)
    self.actor.decoder._last_Ev = torch.stack(_last_Ev)
    self.actor.decoder._last_Eh = torch.stack(_last_Eh)

    v = torch.stack(v)
    h = torch.stack(h)
    return v, h
  
  def learn(self, nodes, rewards):
    """
    @brief update the parameters of actor and critic via REINFORECE algo.
    @param nodes: [#batch, #num_nodes, 2] in dtype float
    @param rewards: reward of an episode in a batch. [#num_batch]
    @note the env has returned 'done' signal
    """
    self.actor_optimizer.zero_grad()
    self.critic_optimizer.zero_grad()

    # cal log(p_res) for each trajectory in the batch
    u0 = self.log_probs_buf['u0'] # [#num_batch]
    u = torch.stack(self.log_probs_buf['u']).transpose(1, 0)  # [#num_batch, #num_nodes - 1]
    ws = torch.stack(self.log_probs_buf['ws']).transpose(1, 0)  # [#num_batch, #num_nodes - 1]
    
    # p_res: [#num_batch]
    p_res = torch.add(u, ws).sum(dim=-1).add(u0)

    with torch.no_grad():
      baselines = self.critic(nodes)
    j = (baselines - rewards) * p_res
    j = -1.0 * j.mean() # the j should be larger after a backward process.

    baselines = self.critic(nodes)
    loss_critic = self.loss_fn_critic(baselines, rewards)
    
    j.backward()
    loss_critic.backward()
    self.actor_optimizer.step()
    self.critic_optimizer.step()

    # Finally, reset the buf
    self.log_probs_buf = {
      "u0": None,
      "u": [],
      "ws": []
    }

