"""
@brief: Network model of Actor and Critic
@author: Zhengyang Lyu
@date: 2022.8.30
"""

import torch
import torch.nn as nn
import pdb

D = 128

class EncoderLayer(nn.Module):
  def __init__(self, num_heads=16, dim_per_head=16, dim=D, h_dim=512) -> None:
    super(EncoderLayer, self).__init__()
    self.num_heads = num_heads
    self.dim_per_head = dim_per_head  # seem no need
    self.dim = dim
    self.h_dim = h_dim

    self.query = nn.Linear(self.dim, self.dim)
    self.key = nn.Linear(self.dim, self.dim)
    self.value = nn.Linear(self.dim, self.dim)
    self.attention = nn.MultiheadAttention(
                        embed_dim=self.dim,
                        num_heads=self.num_heads,
                        batch_first=True
                        )
    
    self.feed_foward = nn.Sequential(
      nn.Linear(self.dim, self.h_dim),
      nn.ReLU(),
      nn.Linear(self.h_dim, self.dim)
    )

    self.norm = nn.BatchNorm1d(num_features=self.dim)
  
  def forward(self, x):
    """
    @param x: [#batch_size, #num_nodes, dim=128]
    """
    batch_size = x.shape[0]
    num_nodes = x.shape[1]
    q = self.query(x)
    k = self.key(x)
    v = self.value(x)
    x1, _ = self.attention(query=q, key=k, value=v, need_weights=False)
    x = x.add(x1)
    # x = self.norm(x.reshape(-1, x.shape[-1])).reshape(batch_size, num_nodes, -1)
    x = x.transpose(1, 2)
    x = self.norm(x)
    x = x.transpose(1, 2)

    x1 = self.feed_foward(x)
    x = x.add(x1)
    # x = self.norm(x.reshape(-1, x.shape[-1])).reshape(batch_size, num_nodes, -1)
    x = x.transpose(1, 2)
    x = self.norm(x)
    x = x.transpose(1, 2)
    return x

class Encoder(nn.Module):
  def __init__(self, dim=D, N=3) -> None:
    super(Encoder, self).__init__()
    self.dim = dim
    self.N = N
    self.W_emb = nn.Linear(2, self.dim, bias=False)  # get E_0
    self.encoder_layers = nn.Sequential()
    for i in range(self.N):
      self.encoder_layers.add_module('{}_{}'.format(EncoderLayer.__name__, i), 
                                     EncoderLayer(num_heads=16))
    self.norm = nn.BatchNorm1d(num_features=self.dim)

  def forward(self, x):
    """
    @param x: [#batch, #num_nodes, 2] in dtype float
    """
    batch_size = x.shape[0]
    num_nodes = x.shape[1]
    x = self.W_emb(x)
    x = x.transpose(1, 2)
    x = self.norm(x)
    x = x.transpose(1, 2)
    """
    dim = x.shape[-1]
    x = x.view(-1, dim)
    x = self.norm(x)
    x = x.reshape(batch_size, num_nodes, -1)
    """
    x = self.encoder_layers(x)
    return x

class PTM(nn.Module):
  # PTM: PoinTing Mechanism
  def __init__(self, dim_e=D, dim_q=360) -> None:
    super(PTM, self).__init__()
    self.dim_e = dim_e
    self.dim_q = dim_q
    self.W_e = nn.Linear(dim_e, dim_q, bias=False)
    self.W_q = nn.Linear(dim_q, dim_q, bias=False)
    self.W_g = nn.Sequential(
      nn.Tanh(),
      nn.Linear(dim_q, 1, bias=False),
    )
    self.C = 10.0

  def forward(self, e, q, mask=None, need_l=False):
    """
    @param e: [#num_batch, #num_nodes, D]
    @param q: [#num_batch, 360]
    @param mask: [#num_batch, #num_nodes], with masked points set to True and the reset are False
                 set visited points True
    @return [#num_batch, #num_nodes], [#num_batch, #num_nodes]
    """
    e = self.W_e(e)
    q = self.W_q(q)
    # e: [#num_batch, #num_nodes, 360]
    # q: [#num_batch,           , 360]
    # for each batch, perform e + q (q is broadcasted in to #num_nodes)
    e = e.transpose_(0, 1).add(q).transpose(0, 1)
    e = self.W_g(e) # get l
    e = e.squeeze()
    l = e.clone()
    if mask != None:
      # points be masked is set to be -INF
      e = torch.where(mask == False, e, -torch.inf)
    e = torch.tanh(e)
    e = e.mul(self.C)
    p = nn.functional.softmax(e, dim=1)
    if need_l:
      return p, l
    return p, None

class EdgeGen(nn.Module):
  """
  @brief Generate query for t-th action
  """
  def __init__(self, dim_e=D, dim_q=360) -> None:
    super(EdgeGen, self).__init__()
    self.dim_e = dim_e
    self.dim_q = dim_q
    self.W_u = nn.Linear(self.dim_e, self.dim_q, bias=False)
    self.W_w = nn.Linear(self.dim_e, self.dim_q, bias=False)
    self.W_v = nn.Linear(self.dim_e, self.dim_q, bias=False)
    self.W_h = nn.Linear(self.dim_e, self.dim_q, bias=False)

  def forward(self, u, w, v, h):
    """
    @param u, w, v, h: [#num_batch, D]
    """
    # edge = self.W_u(u) + self.W_w(w) + self.W_v(v) + self.W_h(h)
    """
    in t = 1, last_u == last_v = u0, last_w = None, last_h = None
    """
    edge = self.W_u(u) + self.W_v(v)
    if not w == None:
      edge += self.W_w(w)
    if not h == None:
      edge += self.W_h(h)
    return edge

class SubTreeGen(nn.Module):
  def __init__(self, dim_q=360) -> None:
    super(SubTreeGen, self).__init__()
    self.dim_q = dim_q
    self.W_edge = nn.Linear(self.dim_q, self.dim_q, bias=False)

  def forward(self, last_subtree, cur_edge):
    if torch.is_tensor(last_subtree):
      cur_subtree = torch.max(last_subtree, self.W_edge(cur_edge))
    else:
      # last_subtree is set to 0 at first iteration
      cur_subtree = torch.relu(self.W_edge(cur_edge))
    return cur_subtree

class QGen(nn.Module):
  def __init__(self, dim_e=D, dim_q=360) -> None:
    super(QGen, self).__init__()
    self.W_5 = nn.Linear(dim_e, dim_q, bias=False)

  def forward(self, last_edge, last_subtree, cur_u=None):
    if cur_u == None:
      cur_q = torch.relu(last_edge + last_subtree)
    else:
      # ERROR! where is u?
      tmp = self.W_5(cur_u)
      cur_q = torch.relu(last_edge + last_subtree + tmp)
    return cur_q

class EPTM(nn.Module):
  # EPTM: Extended PoinTing Mechanism
  def __init__(self, dim_e=D, dim_q=360) -> None:
    super(EPTM, self).__init__()
    self.PTM_0 = PTM(dim_e=dim_e, dim_q=dim_q)
    self.PTM_1 = PTM(dim_e=dim_e, dim_q=dim_q)
    self.C = 10.0

  def forward(self, e, q, mask=None):
    """
    @param e: [#num_batch, #num_nodes, D]
    @param q: [#num_batch, 360]
    @param mask: [#num_batch, #num_nodes], with masked points set to True and the reset are False
                 set unvisited points True
    @return [#num_batch, 2 * #num_nodes]
    @note result[i, j]: in i-th sample the batch, 
        if j < #num_nodes: the prob of s = 0, and w = j
        else: the prob of s = 0, and w = j - #num_nodes
    """
    _, l_0 = self.PTM_0(e, q, need_l=True)  # [#num_batch, #num_nodes]
    _, l_1 = self.PTM_1(e, q, need_l=True)  # [#num_batch, #num_nodes]
    l = torch.stack([l_0, l_1], dim=1)  # [#num_batch, 2, #num_nodes]
    if mask is not None:
      if mask.dim() == 2:
        mask = torch.stack([mask, mask], dim=1)
      l = torch.where(mask == False, l, -torch.inf)
    l = torch.tanh(l).mul(self.C)
    batch_size = l.shape[0]
    p4ws = torch.softmax(l.reshape(batch_size, -1), dim=-1)
    return p4ws

class Decoder(nn.Module):
  def __init__(self, dim_e=D, dim_q=360) -> None:
    super(Decoder, self).__init__()
    self.dim_e = dim_e
    self.dim_q = dim_q
    self.edge_gen = EdgeGen(dim_e, dim_q)
    self.subtree_gen = SubTreeGen(dim_q)
    self.ptm = PTM(dim_e, dim_q)
    self.eptm = EPTM(dim_e, dim_q)
    self.q_gen = QGen(dim_e, dim_q)

    self.ptm_0 = PTM(dim_e=dim_e, dim_q=dim_q)  # for generating u0

    self._last_subtree = 0 # should be a vector, but set to 0 at first
    self._last_edge = None
    self._last_Eu = None
    self._last_Ew = None
    self._last_Ev = None
    self._last_Eh = None
  
  def _get_u0_probs(self, e):
    """
    @brief get a probability distribution of u0
    @param e: [#batch_size, #num_nodes, self.dim_e]
    @return probs [#batch_size, #num_nodes]
    """
    q = torch.zeros(e.shape[0], self.dim_q, device=e.device)
    start_node_probs, _ = self.ptm_0(e, q)
    return start_node_probs
  
  def get_u_probs(self, e, mask_visited=None, mask_unvisited=None):
    """
    @param e: input embeddings for all nodes, [#num_batch, #num_nodes, D]
    @note  define subtree(t=0) = 0
    @return probability of u, ws
            u, w: [#num_batch, #num_nodes]
            s: [#num_batch, #num_nodes, 2]
    """
    self._last_edge = self.edge_gen(self._last_Eu, self._last_Ew, self._last_Ev, self._last_Eh)
    self._last_subtree = self.subtree_gen(self._last_subtree, self._last_edge)

    cur_q4u = self.q_gen(self._last_edge, self._last_subtree)
    u_probs, _ = self.ptm(e, cur_q4u, mask=mask_visited)
    return u_probs

  def get_ws_probs(self, e, E_u, mask_visited=None, mask_unvisited=None):
    """
    @param u: is the node index choiced
    """
    cur_q4w = self.q_gen(self._last_edge, self._last_subtree, E_u)
    ws_probs = self.eptm(e, cur_q4w, mask=mask_unvisited)

    return ws_probs
